@@ -14,7 +14,7 @@ import collections
 import datetime
 import os
 
-from typing import Any, Dict, List, Optional, Set, cast
+from typing import Any, Callable, Dict, List, Optional, Set, cast
 
 import packaging.version
 import semantic_version
@@ -137,14 +137,18 @@ class ChangesBase(metaclass=abc.ABCMeta):
         self.ancestor = self.data.get('ancestor')
 
     @abc.abstractmethod
-    def update_plugins(self, plugins: List[PluginDescription]) -> None:
+    def update_plugins(self, plugins: List[PluginDescription],
+                       allow_removals: Optional[bool]) -> None:
         """
         Update plugin descriptions, and remove plugins which are not in the provided list
         of plugins.
         """
 
     @abc.abstractmethod
-    def update_fragments(self, fragments: List[ChangelogFragment]) -> None:
+    def update_fragments(self, fragments: List[ChangelogFragment],
+                         load_extra_fragments: Optional[
+                             Callable[[str], List[ChangelogFragment]]] = None
+                         ) -> None:
         """
         Update fragment contents, and remove fragment contents which are not in the provided
         list of fragments.
@@ -346,11 +350,17 @@ class ChangesMetadata(ChangesBase):
 
             self.known_fragments |= set(config.get('fragments', []))
 
-    def update_plugins(self, plugins: List[PluginDescription]) -> None:
+    def update_plugins(self, plugins: List[PluginDescription],
+                       allow_removals: Optional[bool]) -> None:
         """
         Update plugin descriptions, and remove plugins which are not in the provided list
         of plugins.
         """
+        if not allow_removals:
+            # We only remove here, we don't update, since we only keep references
+            # to the plugins.
+            return
+
         valid_plugins = collections.defaultdict(set)
 
         for plugin in plugins:
@@ -378,7 +388,10 @@ class ChangesMetadata(ChangesBase):
                     self.known_plugins -= set(
                         '%s/%s' % (plugin_type, plugin) for plugin in invalid_plugins)
 
-    def update_fragments(self, fragments: List[ChangelogFragment]) -> None:
+    def update_fragments(self, fragments: List[ChangelogFragment],
+                         load_extra_fragments: Optional[
+                             Callable[[str], List[ChangelogFragment]]] = None
+                         ) -> None:
         """
         Update fragment contents, and remove fragment contents which are not in the provided
         list of fragments.
@@ -531,7 +544,8 @@ class ChangesData(ChangesBase):
 
             self.known_fragments |= set(config.get('fragments', []))
 
-    def update_plugins(self, plugins: List[PluginDescription]) -> None:
+    def update_plugins(self, plugins: List[PluginDescription],
+                       allow_removals: Optional[bool]) -> None:
         """
         Update plugin descriptions, and remove plugins which are not in the provided list
         of plugins.
@@ -546,36 +560,59 @@ class ChangesData(ChangesBase):
                 invalid_module_names = set(
                     module['name'] for module in config['modules']
                     if module['name'] not in valid_plugins['module'])
-                config['modules'] = [
-                    self._create_plugin_entry(valid_plugins['module'][module['name']])
-                    for module in config['modules']
-                    if module['name'] not in invalid_module_names]
-                self.known_plugins -= set(
-                    'module/%s' % module_name for module_name in invalid_module_names)
+                if allow_removals:
+                    config['modules'] = [
+                        self._create_plugin_entry(valid_plugins['module'][module['name']])
+                        for module in config['modules']
+                        if module['name'] not in invalid_module_names]
+                    self.known_plugins -= set(
+                        'module/%s' % module_name for module_name in invalid_module_names)
+                else:
+                    config['modules'] = [
+                        self._create_plugin_entry(valid_plugins['module'][module['name']])
+                        if module['name'] not in invalid_module_names else
+                        module
+                        for module in config['modules']]
 
             if 'plugins' in config:
                 for plugin_type in config['plugins']:
                     invalid_plugin_names = set(
                         plugin['name'] for plugin in config['plugins'][plugin_type]
                         if plugin['name'] not in valid_plugins[plugin_type])
-                    config['plugins'][plugin_type] = [
-                        self._create_plugin_entry(valid_plugins[plugin_type][plugin['name']])
-                        for plugin in config['plugins'][plugin_type]
-                        if plugin['name'] not in invalid_plugin_names]
-                    self.known_plugins -= set(
-                        '%s/%s' % (plugin_type, plugin_name)
-                        for plugin_name in invalid_plugin_names)
+                    if allow_removals:
+                        config['plugins'][plugin_type] = [
+                            self._create_plugin_entry(valid_plugins[plugin_type][plugin['name']])
+                            for plugin in config['plugins'][plugin_type]
+                            if plugin['name'] not in invalid_plugin_names]
+                        self.known_plugins -= set(
+                            '%s/%s' % (plugin_type, plugin_name)
+                            for plugin_name in invalid_plugin_names)
+                    else:
+                        config['plugins'][plugin_type] = [
+                            self._create_plugin_entry(valid_plugins[plugin_type][plugin['name']])
+                            if plugin['name'] not in invalid_plugin_names else
+                            plugin
+                            for plugin in config['plugins'][plugin_type]]
 
-    def update_fragments(self, fragments: List[ChangelogFragment]) -> None:
+    def update_fragments(self, fragments: List[ChangelogFragment],
+                         load_extra_fragments: Optional[
+                             Callable[[str], List[ChangelogFragment]]] = None
+                         ) -> None:
         """
         Update fragment contents, and remove fragment contents which are not in the provided
         list of fragments.
 
-        Must only be called if ``keep_fragments`` is set to ``True`` in the configuration.
+        Must only be called if ``keep_fragments`` is set to ``True`` in the configuration,
+        or if ``load_extra_fragments`` is provided.
         """
-        assert self.config.keep_fragments
-        valid_fragments = {fragment.name: fragment for fragment in fragments}
-        for _, config in self.releases.items():
+        assert self.config.keep_fragments or load_extra_fragments
+        valid_fragments_global = {fragment.name: fragment for fragment in fragments}
+        for version, config in self.releases.items():
+            valid_fragments = dict(valid_fragments_global)
+            if load_extra_fragments:
+                extra_fragments = load_extra_fragments(version)
+                valid_fragments.update({fragment.name: fragment for fragment in extra_fragments})
+
             config.pop('changes', None)
 
             if 'fragments' in config:
@@ -780,20 +817,15 @@ def add_release(config: ChangelogConfig,  # pylint: disable=too-many-arguments
     changes.save()
 
     if not config.keep_fragments:
-        for fragment in fragments_added:
-            fragment.remove()
-
-
-def refresh_changelog(config: ChangelogConfig,
-                      changes: ChangesBase,
-                      plugins: List[PluginDescription],
-                      fragments: List[ChangelogFragment]):
-    """
-    Refresh plugin information and (depending on config) changelog fragments in ``changes``.
-    """
-    changes.update_plugins(plugins)
-    if config.keep_fragments:
-        changes.update_fragments(fragments)
-    else:
-        LOGGER.warning('Cannot refresh changelog fragments, as keep_fragments is false')
-    changes.save()
+        archive_path_template = config.archive_path_template
+        if archive_path_template is None:
+            # Actually remove fragments
+            for fragment in fragments_added:
+                fragment.remove()
+        else:
+            # Move fragments
+            archive_path = os.path.join(
+                config.paths.base_dir, archive_path_template.format(version=version))
+            os.makedirs(archive_path, exist_ok=True)
+            for fragment in fragments_added:
+                fragment.move_to(archive_path)
